/*
 *  ether_windows.cpp - Ethernet device driver
 *
 *  Basilisk II (C) 1997-1999 Christian Bauer
 *
 *  Ported to Windows by Lauri Pesonen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <process.h>

#include "sysdeps.h"
#include "main.h"
#include "prefs.h"
#include "macos_util.h"
#include "ether.h"
#include "ether_defs.h"
#include <windowsx.h>
#include <winioctl.h>
// #include <ntddndis.h>
#include "main_windows.h"
#include "desktop_windows.h"
#include "b2ether\inc\ntddpack.h"
#include "b2ether\inc\b2ether_hl.h"


#define DEBUG 0
#include "debug.h"

#define MONITOR 0


const uint32 MSG_CLEANUP = 'clea';      // Remove all protocols
const uint32 MSG_ADD_MULTI = 'addm';    // Add multicast address
const uint32 MSG_DEL_MULTI = 'delm';    // Delete multicast address
const uint32 MSG_ATTACH_PH = 'atph';    // Attach protocol handler
const uint32 MSG_DETACH_PH = 'deph';    // Detach protocol handler
const uint32 MSG_WRITE = 'writ';				// Write packet


// #define B2E_NAME "PACKET"
#define B2E_NAME "B2ETHER"

static HANDLE adapter_file = 0;
static CONTROL_BLOCK Adapter;
static CRITICAL_SECTION csection;
static HANDLE hthread = NULL;
static unsigned int thread_id = 0;

static unsigned int ether_thread_func(void *arg);


static LPADAPTER PacketOpenAdapter( char *name )
{
  LPADAPTER lpAdapter = 0;
  int rc;
  char dnamebuf[64];
  char dnamel[64];

  D(bug("PacketOpenAdapter\n"));

  lpAdapter = (LPADAPTER)GlobalAllocPtr( GMEM_MOVEABLE|GMEM_ZEROINIT, sizeof(ADAPTER) );
  if(!lpAdapter) {
    D(bug("PacketOpenAdapter GlobalAlloc Failed\n"));
    return(0);
  }

  // "\\.\b2ether_<ndis>"
  wsprintf( lpAdapter->SymbolicLink, TEXT("\\\\.\\%s_%s"), B2E_NAME, name );

  if ( QueryDosDevice(&lpAdapter->SymbolicLink[4], dnamebuf, sizeof(dnamebuf)) == 0) {
    rc = GetLastError();
    if ( rc == ERROR_FILE_NOT_FOUND ) {
      sprintf(dnamel,"\\Device\\%s_%s",B2E_NAME,name);
      if ( DefineDosDevice(DDD_RAW_TARGET_PATH, &lpAdapter->SymbolicLink[4], dnamel) == 0 ) {
        // clean up
        D(bug("DefineDosDevice failed in PacketOpenAdapter\n"));
        goto cleanup;
        return( NULL );
      }
    } else {
      // clean up
      D(bug("QueryDosDevice failed (%d) in PacketOpenAdapter\n",rc));
      goto cleanup;
      return( NULL );
    }
  }

  lpAdapter->hFile = CreateFile(
        lpAdapter->SymbolicLink,
        GENERIC_WRITE | GENERIC_READ,
        0,
        NULL,
        OPEN_EXISTING, // CREATE_ALWAYS,
        FILE_FLAG_OVERLAPPED,
        0
        );

  if (lpAdapter->hFile != INVALID_HANDLE_VALUE)
    return lpAdapter;

  D(bug("PacketOpenAdapter Could not open adapter %s.\n",lpAdapter->SymbolicLink));

cleanup:
  if(lpAdapter) GlobalFreePtr( lpAdapter );
  GlobalFreePtr( lpAdapter );
  return( NULL );
}

static BOOLEAN StartPacketDriver( LPTSTR ServiceName )
{
  BOOLEAN  Status;
  SC_HANDLE  SCManagerHandle;
  SC_HANDLE  SCServiceHandle;

  SCManagerHandle = OpenSCManager(
                    NULL,
                    NULL,
                    SC_MANAGER_ALL_ACCESS);

  if (SCManagerHandle==NULL) {
    D(bug("Could not open service control manager"));
    return FALSE;
  }
  SCServiceHandle = OpenService( SCManagerHandle, ServiceName, SERVICE_START );
  if (SCServiceHandle == NULL) {
    D(bug("Could not open service %s",ServiceName));
    return FALSE;
  }
  Status = StartService( SCServiceHandle, 0, NULL );
  if (!Status) {
    if (GetLastError()==ERROR_SERVICE_ALREADY_RUNNING) {
      D(bug("Service %s is already running.\n",ServiceName ));
      return TRUE;
    }
  }
  return(FALSE);
}

static LPPACKET PacketAllocatePacket( LPADAPTER AdapterObject )
{
  LPPACKET p;

  p = (LPPACKET)GlobalAllocPtr( GMEM_MOVEABLE|GMEM_ZEROINIT, sizeof(PACKET) );
  if(!p) {
    D(("PacketAllocateSendPacket: GlobalAlloc Failed\n"));
    return(0);
  }
  p->OverLapped.hEvent = CreateEvent(0,0,0,0);
  if (!p->OverLapped.hEvent) {
    D(("PacketAllocateSendPacket: CreateEvent Failed\n"));
    GlobalFreePtr(p);
    return(0);
  }
  return(p);
}

static void PacketFreePacket( LPPACKET p )
{
  CloseHandle(p->OverLapped.hEvent);
  GlobalFreePtr(p);
}

static void PacketInitPacket( LPPACKET p, PVOID Buffer, UINT Length )
{
  p->Buffer=Buffer;
  p->Length=Length;
}

static BOOLEAN PacketSendPacket(
  LPADAPTER   AdapterObject,
  LPPACKET    lpPacket,
  BOOLEAN     Sync
)
{
  BOOLEAN    Result;
  DWORD      BytesTransfered;

  lpPacket->OverLapped.Offset = 0;
  lpPacket->OverLapped.OffsetHigh = 0;

  if (!ResetEvent(lpPacket->OverLapped.hEvent)) {
    return FALSE;
  }
  Result = WriteFile(
            AdapterObject->hFile,
            lpPacket->Buffer,
            lpPacket->Length,
            &BytesTransfered,
            &lpPacket->OverLapped
            );
  if (Sync) {
    Result = GetOverlappedResult(
               AdapterObject->hFile,
               &lpPacket->OverLapped,
               &BytesTransfered,
               TRUE
               );
  } else {
    Result=TRUE;
  }
  return Result;
}

/*
 *  Initialization
 */

void EtherInit(void)
{
  memset( &Adapter, 0, sizeof(Adapter) );

  InitializeCriticalSection( &csection );

	if(win_os != VER_PLATFORM_WIN32_NT) 
    return;

  if (PrefsFindString("ether") == NULL)
    return;

  strcpy( Adapter.AdapterName, PrefsFindString("ether") );

  if(!StartPacketDriver(B2E_NAME)) {
    D(bug("Could not start BasiliskII ethernet driver \"%s\".",B2E_NAME));
    return;
  }

  Adapter.hFile = PacketOpenAdapter( Adapter.AdapterName );
  if (Adapter.hFile == NULL) {
    D(bug("Failed to open adapter \"%s\".",Adapter.AdapterName));
    // ULONG NameLength = 64;
    // PacketGetAdapterNames( Adapter.AdapterName, &NameLength );
    // Show a list of available adapters
    return;
  }

  if (Adapter.hFile == NULL)
    return;

	hthread = (HANDLE)_beginthreadex( 0, 0, ether_thread_func, 0, 0, &thread_id );
  if (!hthread) {
    D(bug("Failed to create ethernet thread"));
    EtherExit();
    return;
  }

  // preliminary test
  // these will go away soon

  Adapter.BufferSize = 1514;

  Adapter.hMem   = GlobalAlloc( GMEM_MOVEABLE, Adapter.BufferSize );
  Adapter.lpMem  = (LPBYTE)GlobalLock( Adapter.hMem );
  Adapter.hMem2  = GlobalAlloc( GMEM_MOVEABLE, Adapter.BufferSize );
  Adapter.lpMem2 = (LPBYTE)GlobalLock( Adapter.hMem2 );
}

/*
 *  Deinitialization
 */

void EtherExit(void)
{
  D(bug("EtherExit\n"));
  if(Adapter.hFile) {
    CloseHandle(Adapter.hFile->hFile);
    GlobalFreePtr(Adapter.hFile);
    Adapter.hFile = 0;
  }
  if(Adapter.lpMem2) {
    GlobalUnlock( Adapter.hMem2 );
    Adapter.lpMem2 = 0;
  }
  if(Adapter.hMem2) {
    GlobalFree( Adapter.hMem2 );
    Adapter.hMem2 = 0;
  }
  if(Adapter.lpMem) {
    GlobalUnlock( Adapter.hMem );
    Adapter.lpMem = 0;
  }
  if(Adapter.hMem) {
    GlobalFree( Adapter.hMem );
    Adapter.hMem = 0;
  }
  memset( &Adapter, 0, sizeof(Adapter) ); // preparing for B2 security (just kidding)
  DeleteCriticalSection( &csection );
}


/*
 *  Send message to network process
 */

static int16 send_to_proc(uint32 what, uint32 pointer = 0, uint16 type = 0)
{
  D(bug("sending %08lx to net_proc\n", what));
  return noErr;
}

/*
 *  Reset
 */

void EtherReset(void)
{
  DWORD BytesReturned;
  BOOL result;

  D(bug("EtherReset\n"));

  // Remove all protocols
  if (net_open)
    send_to_proc(MSG_CLEANUP);

  if(Adapter.hFile && Adapter.hFile->hFile) {
    result = DeviceIoControl(
        Adapter.hFile->hFile,
        (DWORD)IOCTL_PROTOCOL_RESET,
        NULL,
        0,
        NULL,
        0,
        &BytesReturned,
        NULL
        );
    D(bug("EtherReset returned %d\n",(int)result));
  }
}


/*
 *  Add multicast address
 */

int16 ether_add_multicast(uint32 pb)
{
  return noErr;
}


/*
 *  Delete multicast address
 */

int16 ether_del_multicast(uint32 pb)
{
  return noErr;
}


/*
 *  Attach protocol handler
 */

int16 ether_attach_ph(uint16 type, uint32 handler)
{
  return noErr;
}


/*
 *  Detach protocol handler
 */

int16 ether_detach_ph(uint16 type)
{
  return noErr;
}


/*
 *  Transmit raw ethernet packet
 */

// First test: synchronous...no thread

int16 ether_write(uint32 wds)
{
  send_to_proc(MSG_WRITE);
  return 1; // Command in progress

  /*
  bool ret;
  LPPACKET  p;

  p = PacketAllocatePacket( Adapter.hFile );
  if(!p) return( controlErr );
  PacketInitPacket( p, Buffer, Length );
  ret = PacketSendPacket( Adapter.hFile, p, TRUE );
  PacketFreePacket( p );

  return ret ? noErr : controlErr;
  */
}


/*
 *  Ethernet interrupt - activate deferred tasks to call IODone or protocol handlers
 */

void EtherInterrupt(void)
{
  D(bug("EtherIRQ\n"));
}

static unsigned int ether_thread_func(void *arg)
{
	set_desktop();

  D(bug("ether_thread_func running"));


  D(bug("ether_thread_func exit"));

  hthread = NULL;
  thread_id = 0;

	_endthreadex( 0 );

  return(0);
}
